
/*
   helmath3.c
   
   Distribution generation routines for helicon.
   
   Author: Mark E. Nash
   Date:   August, 1988
*/

#include	<stdio.h>
#include	<stdlib.h>
#include	<malloc.h>
#include	<math.h>
#include	<string.h>
#include	<conio.h>
#include	"helimacs.h"
#include	"helimess.h"
#include	"helstruc.h"
#include	"helfdef.h"


/* externals */
extern char msg[];				/* in heliglob.h */


int rectgen(x, pa, norm, fv)
	double x;
	double pa[MAXPARS+1][NUMDIST+1];
	unsigned norm;
	double *fv;
/*
   rectgen() generates a point from the rectangular d.f.
   
   Entry: x is the domain value.
          The other parameter values are set in the parameter array pa[][].
	  norm = TRUE for normalized calculation.
	  
   Exit:  Returns TRUE if point is valid,
	  fv is the returned domain value.
*/
	{
	double	a, b, h;	/* parameters */
	
	
	a = pa[APAR][RECT];
	b = pa[BPAR][RECT];
	h = pa[MAG][RECT];
	
	if (x >= a && x <= b)
		*fv = 1.0 / (b - a);
	else
		*fv = 0.0;
	
	if (norm == FALSE)
		*fv *= h;
	
	return(TRUE);
	}
	
	
	
int trigen(x, pa, norm, fv)
	double x;
	double pa[MAXPARS+1][NUMDIST+1];
	unsigned norm;
	double *fv;
/*
   trigen() generates a point from the triangular d.f.
   
   Entry: x is the domain value.
          The other parameter values are set in the parameter array pa[][].
	  norm = TRUE for normalized calculation.
	  
   Exit:  Returns TRUE if point is valid,
	  fv is the returned domain value.
*/
	{
	double	a, b, m, h;	/* parameters */
	
	
	a = pa[APAR][TRI];
	b = pa[BPAR][TRI];
	m = pa[MPAR][TRI];
	h = pa[MAG][TRI];
	
	if (x < a || x > b)
		*fv = 0.0;
	else
		if (x < m)
			*fv = 2.0 * (x-a) / ((m-a) * (b-a));
		else
			*fv = 2.0 * (b-x) / ((b-m) * (b-a));
	
	if (norm == FALSE)
		*fv *= h;
	
	return(TRUE);
	}
	
	
	
int normgen(x, pa, norm, fv)
	double x;
	double pa[MAXPARS+1][NUMDIST+1];
	unsigned norm;
	double *fv;
/*
   normgen() generates a point from the normal d.f.
   
   Entry: x is the domain value.
          The other parameter values are set in the parameter array pa[][].
	  norm = TRUE for normalized calculation.
	  
   Exit:  Returns TRUE if point is valid,
	  fv is the returned domain value.
*/
	{
	double	a, b, h;	/* parameters */
	
	
	a = pa[NMEAN][NOR];
	b = pa[NSDEV][NOR];
	h = pa[MAG][NOR];
	
	*fv = (1.0 / (sqrt(2.0*PI)*b)) * exp((-pow((x-a), 2.0)) / (2.0 * 
								pow(b, 2.0)));
	
	if (norm == FALSE)
		*fv *= h;
	
	return(TRUE);
	}



int gamgen(x, pa, norm, fv)
	double x;
	double pa[MAXPARS+1][NUMDIST+1];
	unsigned norm;
	double *fv;
/*
   gamgen() generates a point from the gamma d.f.
   
   Entry: x is the domain value.
          The other parameter values are set in the parameter array pa[][].
	  norm = TRUE for normalized calculation.
	  
   Exit:  Returns TRUE if point is valid,
	  fv is the returned domain value.
*/
	{
	double	c, r, h;	/* parameters */
	
	
/* exit if not in range */
	*fv = 0.0;
	if (x < 0.0)
		return(FALSE);
		
	c = pa[CPAR][GAM];
	r = pa[RPAR][GAM];
	h = pa[MAG][GAM];
	
	*fv = (pow(c, r) / gamma(r)) * pow(x, r-1.0) * exp(-c*x);
	
	if (norm == FALSE)
		*fv *= h;
	
	return(TRUE);
	}



int pargen(x, pa, norm, fv)
	double x;
	double pa[MAXPARS+1][NUMDIST+1];
	unsigned norm;
	double *fv;
/*
   pargen() generates a point from the pareto d.f.
   
   Entry: x is the domain value.
          The other parameter values are set in the parameter array pa[][].
	  norm = TRUE for normalized calculation.
	  
   Exit:  Returns TRUE if point is valid,
	  fv is the returned domain value.
*/
	{
	double	a, b, h;	/* parameters */
	
	
/* exit if not in range */
	*fv = 0.0;
	if (x <= b)
		return(FALSE);
		
	a = pa[APAR][PAR];
	b = pa[BPAR][PAR];
	h = pa[MAG][PAR];
	
	*fv = a * pow(b, a) / pow(x, a+1.0); 
	
	if (norm == FALSE)
		*fv *= h;
	
	return(TRUE);
	}



int fgen(x, pa, norm, fv)
	double x;
	double pa[MAXPARS+1][NUMDIST+1];
	unsigned norm;
	double *fv;
/*
   fgen() generates a point from the F d.f.
   
   Entry: x is the domain value.
          The other parameter values are set in the parameter array pa[][].
	  norm = TRUE for normalized calculation.
	  
   Exit:  Returns TRUE if point is valid,
	  fv is the returned domain value.
*/
	{
	double	m, n, h;	/* parameters */
	
	
/* exit if not in range */
	*fv = 0.0;
	if (x < 0.0)
		return(FALSE);
		
	m = pa[FMPAR][FDIST];
	n = pa[FNPAR][FDIST];
	h = pa[MAG][FDIST];
	
	*fv = (gamma((m+n)/2.0) / (gamma(m/2.0) * gamma(n/2.0))) *
	   pow(m/n, m/2.0) * pow(x, (m-2.0)/2.0) / pow(1.0+(m/n)*x, (m+n)/2.0);
 
	
	if (norm == FALSE)
		*fv *= h;
	
	return(TRUE);
	}



int betagen(x, pa, norm, fv)
	double x;
	double pa[MAXPARS+1][NUMDIST+1];
	unsigned norm;
	double *fv;
/*
   betagen() generates a point from the beta d.f.
   
   Entry: x is the domain value.
          The other parameter values are set in the parameter array pa[][].
	  norm = TRUE for normalized calculation.
	  
   Exit:  Returns TRUE if the point is valid,
	  fv is the returned domain value.
*/
	{
	double	a, b, h;	/* parameters */
	
	
/* exit if not in range */
	*fv = 0.0;
	if (x < 0.0 || x > 1.0)
		return(FALSE);
		
	a = pa[APAR][BETA];
	b = pa[BPAR][BETA];
	h = pa[MAG][BETA];
	
	*fv = (gamma(a+b) / (gamma(a)*gamma(b))) * pow(x, a-1.0) * 
							pow(1.0-x, b-1.0);
	
	if (norm == FALSE)
		*fv *= h;
	
	return(TRUE);
	}
	
	
	
int lnorgen(x, pa, norm, fv)
	double x;
	double pa[MAXPARS+1][NUMDIST+1];
	unsigned norm;
	double *fv;
/*
   lnorgen() generates a point from the lognormal d.f.
   
   Entry: x is the domain value.
          The other parameter values are set in the parameter array pa[][].
	  norm = TRUE for normalized calculation.
	  
   Exit:  Returns TRUE if the point is valid,
	  fv is the returned domain value.
*/
	{
	double	m, r, h;	/* parameters */
	
	
/* exit if not in range */
	*fv = 0.0;
	if (x <= 0.0)
		return(FALSE);
		
	m = pa[LMPAR][LOGNOR];
	r = pa[LRPAR][LOGNOR];
	h = pa[MAG][LOGNOR];
	
	*fv = (1.0/(x*sqrt(2.0*PI)*r)) * exp(-pow(log(x)-m, 2.0)/(2.0*
								pow(r, 2.0)));
	
	if (norm == FALSE)
		*fv *= h;
	
	return(TRUE);
	}
	
	
	
int exgen(x, pa, norm, fv)
	double x;
	double pa[MAXPARS+1][NUMDIST+1];
	unsigned norm;
	double *fv;
/*
   exgen() generates a point from the exponential d.f.
   
   Entry: x is the domain value.
          The other parameter values are set in the parameter array pa[][].
	  norm = TRUE for normalized calculation.
	  
   Exit:  Returns TRUE if the point is valid,
	  fv is the returned domain value.
*/
	{
	double	c, h;		/* parameters */
	
	
/* exit if not in range */
	*fv = 0.0;
	if (x < 0.0)
		return(FALSE);
		
	c = pa[CPAR][EXPON];
	h = pa[MAG][EXPON];
	
	*fv = c * exp(-c*x);

	if (norm == FALSE)
		*fv *= h;
	
	return(TRUE);
	}
	
	
	
int dexgen(x, pa, norm, fv)
	double x;
	double pa[MAXPARS+1][NUMDIST+1];
	unsigned norm;
	double *fv;
/*
   dexgen() generates a point from the double exponential d.f.
   
   Entry: x is the domain value.
          The other parameter values are set in the parameter array pa[][].
	  norm = TRUE for normalized calculation.
	  
   Exit:  Returns TRUE if the point is valid,
	  fv is the returned domain value.
*/
	{
	double	a, b, h;	/* parameters */
	
	
	a = pa[APAR][DEXPON];
	b = pa[BPAR][DEXPON];
	h = pa[MAG][DEXPON];
	
	*fv = (1.0/(2.0*b)) * exp(-fabs(x-a)/b);

	if (norm == FALSE)
		*fv *= h;
	
	return(TRUE);
	}
	
	
	
int gumgen(x, pa, norm, fv)
	double x;
	double pa[MAXPARS+1][NUMDIST+1];
	unsigned norm;
	double *fv;
/*
   gumgen() generates a point from the Gumbel d.f.
   
   Entry: x is the domain value.
          The other parameter values are set in the parameter array pa[][].
	  norm = TRUE for normalized calculation.
	  
   Exit:  Returns TRUE if the point is valid,
	  fv is the returned domain value.
*/
	{
	double	a, b, h;	/* parameters */
	
	
	a = pa[APAR][GUM];
	b = pa[BPAR][GUM];
	h = pa[MAG][GUM];
	
	*fv = exp(-exp(-(x-a)/b));

	if (norm == FALSE)
		*fv *= h;
	
	return(TRUE);
	}
	
	
	
int chigen(x, pa, norm, fv)
	double x;
	double pa[MAXPARS+1][NUMDIST+1];
	unsigned norm;
	double *fv;
/*
   chigen() generates a point from the chi-square d.f.
   
   Entry: x is the domain value.
          The other parameter values are set in the parameter array pa[][].
	  norm = TRUE for normalized calculation.
	  
   Exit:  Returns TRUE if the point is valid,
	  fv is the returned domain value.
*/
	{
	double	k, h;		/* parameters */
	
	
/* exit if not in range */
	*fv = 0.0;
	if (x < 0.0)
		return(FALSE);
		
	k = pa[KPAR][CHI];
	h = pa[MAG][CHI];
	
	*fv = (1.0 / gamma(k/2.0)) * pow(1.0/2.0, k/2.0) * pow(x, (k/2.0)-1.0)
							 * exp(-(1.0/2.0) * x);

	if (norm == FALSE)
		*fv *= h;
	
	return(TRUE);
	}
	
	
	
int loggen(x, pa, norm, fv)
	double x;
	double pa[MAXPARS+1][NUMDIST+1];
	unsigned norm;
	double *fv;
/*
   loggen() generates a point from the logistic cumulative d.f.
   
   Entry: x is the domain value.
          The other parameter values are set in the parameter array pa[][].
	  norm = TRUE for normalized calculation.
	  
   Exit:  Returns TRUE if the point is valid,
	  fv is the returned domain value.
*/
	{
	double	a, b, h;		/* parameters */
	
	
	a = pa[APAR][LOGIS];
	b = pa[BPAR][LOGIS];
	h = pa[MAG][LOGIS];
	
	*fv = 1.0 / (1.0 + exp(-(x-a)/b));

	if (norm == FALSE)
		*fv *= h;
	
	return(TRUE);
	}
	
	
	
int tgen(x, pa, norm, fv)
	double x;
	double pa[MAXPARS+1][NUMDIST+1];
	unsigned norm;
	double *fv;
/*
   tgen() generates a point from the t d.f.
   
   Entry: x is the domain value.
          The other parameter values are set in the parameter array pa[][].
	  norm = TRUE for normalized calculation.
	  
   Exit:  Returns TRUE if the point is valid,
	  fv is the returned domain value.
*/
	{
	double	k, h;		/* parameters */
	
	
	k = pa[KPAR][TDIST];
	h = pa[MAG][TDIST];
	
	*fv = (gamma((k+1.0)/2.0) / gamma(k/2.0)) * (1.0 / sqrt(k*PI)) *
				(1.0 / pow(1.0 + pow(x, 2.0)/k, (k+1.0)/2.0));

	if (norm == FALSE)
		*fv *= h;
	
	return(TRUE);
	}
	
	
	
int weigen(x, pa, norm, fv)
	double x;
	double pa[MAXPARS+1][NUMDIST+1];
	unsigned norm;
	double *fv;
/*
   weigen() generates a point from the weibull d.f.
   
   Entry: x is the domain value.
          The other parameter values are set in the parameter array pa[][].
	  norm = TRUE for normalized calculation.
	  
   Exit:  Returns TRUE if the point is valid,
	  fv is the returned domain value.
*/
	{
	double	a, b, h;	/* parameters */
	
	
/* exit if not in range */
	*fv = 0.0;
	if (x <= 0.0)
		return(FALSE);
		
	a = pa[APAR][WEI];
	b = pa[BPAR][WEI];
	h = pa[MAG][WEI];
	
	*fv = a * b * pow(x, b-1.0) * exp(-a*pow(x, b));
	
	if (norm == FALSE)
		*fv *= h;
	
	return(TRUE);
	}
	
	
	
int bingen(x, pa, norm, fv)
	double x;
	double pa[MAXPARS+1][NUMDIST+1];
	unsigned norm;
	double *fv;
/*
   bingen() generates a point from the binomial d.f.
   
   Entry: x is the domain value.
          The other parameter values are set in the parameter array pa[][].
	  norm = TRUE for normalized calculation.
	  
   Exit:  Returns TRUE if point is valid,
	  fv is the returned domain value.
*/
	{
	double	p, n, h;		/* parameters */
	
	
/* exit if not in range */
	*fv = 0.0;
	if (x < 0.0)
		return(FALSE);
		
	p = pa[PPAR][BINOM];
	n = pa[NPAR][BINOM];
	h = pa[MAG][BINOM];
	
	*fv = binco((long)n, (long)x) * pow(p, x) * pow(1.0-p, (n-x));
	
	if (norm == FALSE)
		*fv *= h;
	
	return(TRUE);
	}
	
	

int poigen(x, pa, norm, fv)
	double x;
	double pa[MAXPARS+1][NUMDIST+1];
	unsigned norm;
	double *fv;
/*
   poigen() generates a point from the poisson d.f.
   
   Entry: x is the domain value.
          The other parameter values are set in the parameter array pa[][].
	  norm = TRUE for normalized calculation.
	  
   Exit:  Returns TRUE if point is valid,
	  fv is the returned domain value.
*/
	{
	double	c, h;			/* parameters */
	

/* exit if not in range */
	*fv = 0.0;
	if (x < 0.0)
		return(FALSE);
		
	c = pa[CPAR][POI];
	h = pa[MAG][POI];
	
	*fv = exp(-c) * pow(c, x) / fact((int)x);
	
	if (norm == FALSE)
		*fv *= h;
	
	return(TRUE);
	}
	
	
	
int geogen(x, pa, norm, fv)
	double x;
	double pa[MAXPARS+1][NUMDIST+1];
	unsigned norm;
	double *fv;
/*
   geogen() generates a point from the geometric d.f.
   
   Entry: x is the domain value.
          The other parameter values are set in the parameter array pa[][].
	  norm = TRUE for normalized calculation.
	  
   Exit:  Returns TRUE if point is valid,
	  fv is the returned domain value.
*/
	{
	double	p, h;			/* parameters */
	
	
/* exit if not in range */
	*fv = 0.0;
	if (x < 0.0)
		return(FALSE);
		
	p = pa[PPAR][GEO];
	h = pa[MAG][GEO];
	
	*fv = p * pow((1.0-p), x);
	
	if (norm == FALSE)
		*fv *= h;
	
	return(TRUE);
	}
